from .summarization_step import SummarizationStep
from .translation_step import TranslationStep
from .control_step import ControlStep
from .apicall_step import ApiCallStep
from .validation_step import ValidationStep



from ...models import LLModel
from ..base import Step


class ControlStep(Step):

    name = "control_step"
    description = "selects the required module for execution to solve the task"
    properties = dict(
        task=dict(
            type='string',
            description="input task to solve",
            required=True
        )
    )
    prompt = """
    Your task is to determine which of the steps you need to call at the current moment, based on the available data, previous steps results and the task. \
    If you think that the end result is achieved, then you must call a step called "end_step". \
    In the "end_step" step, put the result that corresponds to the task. \
    Do not use the result of the "validation_step" step as an answer. \
    Ask for clarification if a user task is ambiguous.
    """

    def __init__(self, model: LLModel, steps: dict, steps_specs: list) -> None:
        super(ControlStep, self).__init__(model)
        self.steps = steps
        self.steps_specs = steps_specs

    @property
    def history(self) -> list:
        return self.model.history


    def __call__(self, task: str) -> str:
        response = self.model(content=self.prompt, input_text=task, functions=self.steps, functions_spec=self.steps_specs)
        return response



from ...models import LLModel
from ..base import Step


class ApiCallStep(Step):

    name = "api_call_step"
    description = "gives the necessary api request to the service to solve the task"
    properties = dict(
        task=dict(
            type='string',
            description="input task to solve with api request",
            required=True
        )
    )
    prompt = """
    Ask for clarification if a user task is ambiguous.
    """

    def __init__(self, model: LLModel, endpoints: dict, endpoints_specs: list) -> None:
        super(ApiCallStep, self).__init__(model)
        self.endpoints = endpoints
        self.endpoints_specs = endpoints_specs

    @property
    def history(self) -> list:
        return self.model.history

    def __call__(self, task: str) -> str:
        response = self.model(content=self.prompt, input_text=task, functions=self.endpoints, functions_spec=self.endpoints_specs)
        return response



import abc
from copy import copy
from ..models import LLModel


class Step(abc.ABC):

    def __init__(self, model: LLModel) -> None:
        self.model = model

    @property
    @abc.abstractmethod
    def name(self) -> str:
        pass

    @property
    @abc.abstractmethod
    def description(self) -> str:
        pass

    @property
    @abc.abstractmethod
    def properties(self) -> dict:
        pass

    @property
    def spec(self) -> dict:
        required_list = []
        properties = copy(self.properties)
        for prop, prop_info in properties.items():
            if prop_info.get('required', False):
                required_list.append(prop)
                prop_info.pop('required')
        return dict(
            name=self.name,
            description=self.description,
            parameters=dict(
                type='object',
                properties=properties,
                required=required_list
            )
        )
    
    @property
    @abc.abstractmethod
    def prompt(self) -> str:
        pass
    
    @abc.abstractmethod
    def __call__(self, *args, **kwds) -> str:
        pass



from ..base import Step


class SummarizationStep(Step):

    name = "summarization_step"
    description = "summarize the input text"
    properties = dict(
        text=dict(
            type='string',
            description="input text to summarize",
            required=True
        ),
        action=dict(
            type='string',
            enum=['summarize', 'describe', 'extract main information from'],
            required=True
        ),
        words_limitation=dict(
            type='integer',
            description="maximum words in summary"
        )
    )
    prompt = """
    Your task is to {action} the text. \
    {words_limitation}
    """

    def __call__(self, text: str, action: str, words_limitation=None) -> str:
        prompt = self.prompt
        if words_limitation:
            prompt = prompt.replace('{words_limitation}', f'Do it in at most {words_limitation} words.')
        prompt = prompt.replace('{action}', action)
        response = self.model(content=prompt, input_text=text)
        return response



from ..base import Step


class ValidationStep(Step):

    name = "validation_step"
    description = "verifies that the answer is correct according the given task"
    properties = dict(
        answer=dict(
            type='string',
            description="answer to the task",
            required=True
        ),
        task=dict(
            type='string',
            description="given task",
            required=True
        ),
        context=dict(
            type='string',
            description='task context',
            required=True
        )
    )
    prompt = """
    You must check that the answer is correct for the task according the context. \
    You must admit that there may be a little inaccuracies in the answer. You should still count this answer as correct.
    Print result only by following dict format: {"evaluation": <evaluation_result>, "explanation": <your explanation>} and nothing else. \
    If this answer is correct, then you should replace <evaluation_result> with the exact message "valid" in "" quotes and replace <your explanation> with empty "" quotes. \
    If this answer is not correct, then you must should replace <evaluation_result> with the exact message "invalid" in "" quotes and deduce why the answer is not relevant to the context and place it into <your explanation> in "" quotes. \
    Your <evaluation_result> if it is not empty must be no more than 50 words. \
    """

    def __call__(self, answer: str, task: str, context: str) -> str:
        input_text = str(dict(answer=answer, task=task, context=context))
        response = self.model(content=self.prompt, input_text=input_text)
        return response



from ..base import Step


class StartStep(Step):

    name = "start_step"
    description = "starting scenario flow"
    properties = dict(
        task=dict(
            type='string',
            description="input text",
            required=True
        )
    )
    prompt = """"""

    def __call__(self, task: str) -> str:
        return task



from ..base import Step


class EndStep(Step):

    name = "end_step"
    description = "ending scenario flow"
    properties = dict(
        result=dict(
            type='string',
            description="input text",
            required=True
        )
    )
    prompt = """"""

    def __call__(self, result: str) -> str:
        return result



from .base import LLModel
import openai
import json


class OpenAIModel(LLModel):

    def __init__(self, model_name: str, openai_token: str, history_volume=0, **model_params) -> None:
        super(OpenAIModel, self).__init__()
        openai.api_key = openai_token
        self.model_name = model_name
        self.model_params = model_params
        self._history_volume = history_volume
        self.history = []
    
    def __call__(self, content: str, input_text: str, functions=None, functions_spec=None) -> str:
        if functions:
            response = self._functions(content, input_text, functions=functions, functions_spec=functions_spec)
        else:
            response =  self._completion(content, input_text)
        return response
    
    def save_history(self, *messages):
        if self._history_volume == 0:
            return
        if len(self.history) >= self._history_volume:
            self.history = self.history[1:]
        for message in messages:
            self.history.append(message)
    
    def _completion(self, content: str, input_text: str) -> str:
        messages = [dict(role='system', content=content)]
        for hist in self.history:
            messages.append(dict(role=hist['role'], content=hist['content']))
        messages.append(dict(role='user', content=input_text))
        try:
            response = openai.ChatCompletion.create(
                model=self.model_name,
                messages=messages,
                **self.model_params
            ).get('choices')[0].get('message', dict()).get('content', "")
        except KeyError:
            response = "Key Error [0]. Do not have enough information."
        self.save_history(messages[-1], dict(role='assistant', content=response))
        return response

    def _functions(self, content: str, input_text: str, functions=None, functions_spec=None) -> str:
        messages = [dict(role='system', content=content)]
        for hist in self.history:
            messages.append(dict(role=hist['role'], content=hist['content']))
        messages.append(dict(role='user', content=input_text))
        try:
            response = openai.ChatCompletion.create(
                model=self.model_name,
                messages=messages,
                functions=functions_spec,
                function_call='auto',
                **self.model_params
            ).get('choices')[0].get('message', {})
        except KeyError:
            response = {"error": "Key Error [0]. Do not have enough information."}
        result = {'message': response.get('content', "")}
        function_call = response.get('function_call')
        if function_call:
            function_name = function_call['name']
            function_kwds = json.loads(function_call['arguments'].replace('\n', ''))
            try:
                function_response = functions[function_name](**function_kwds)
            except TypeError as ex:
                function_response = str(ex)
            call_result = {'name': function_name, 'arguments': str(function_kwds)}
            result = {'result': function_response, 'function_call': str(call_result)}
            self.save_history(messages[-1], dict(role='assistant', content=function_response, function_call=str(call_result)))
        else:
            self.save_history(messages[-1], dict(role='assistant', content=response.get('content', "")))
        return result



from ..base import Step


class TranslationStep(Step):

    name = "translation_step"
    description = "translate the input text from one language to another"
    properties = dict(
        text=dict(
            type='string',
            description="input text to translate",
            required=True
        ),
        from_language=dict(
            type='string',
            description="the language from which the text needs to be translated",
            required=True
        ),
        to_language=dict(
            type='string',
            description="the language into which the text needs to be translated",
            required=True
        ),
        style=dict(
            type='string',
            description="translation style",
            required=True
        )
    )
    prompt = """
    Your task is to translate the text from {from_language} to {to_language} in {style} style. \
    You must keep all special characters in the text like: "<", ">", "{", "}", "[", "]".
    """

    def __call__(self, text: str, from_language: str, to_language: str, style: str) -> str:
        prompt = self.prompt
        prompt = prompt.replace('{from_language}', from_language).replace('{to_language}', to_language)
        prompt = prompt.replace('{style}', style)
        response = self.model(content=prompt, input_text=text)
        return response



from enum import Enum
from ..steps import ControlStep
import ast


class ScenarioMode(Enum):

    AUTO = 0
    DEFINED = 1


class Scenario(object):

    def __init__(self) -> None:
        self.steps = dict()
        self.steps_specs = list()
        self.control_step = None
        
    def setup(self, steps: list, needs_control=False, control_model=None):
        self.steps = {step.name: step for step in steps}
        self.steps_specs = [step.spec for step in steps]
        if needs_control:
            self.control_step = ControlStep(control_model, self.steps, self.steps_specs)


    def run(self, task: str, mode: ScenarioMode, flow=None, max_attempts=10):
        scenario_run = ScenarioRun(self, task=task, mode=mode, flow=flow, control_step=self.control_step, max_attempts=max_attempts)
        return scenario_run.execute()



class ScenarioRun:


    def __init__(self, scenario: Scenario, **run_params) -> None:
        self.scenario = scenario
        self.__dict__.update(**run_params)
        self.__execute_switch = {
            ScenarioMode.AUTO: self.run_auto,
            ScenarioMode.DEFINED: self.run_static
        }
        self._history = dict()
        self.result = None

    @property
    def history(self) -> dict:
        return dict(sorted(self._history.items(), key=lambda item: item[1]['id']))

    def execute(self):
        return self.__execute_switch[self.mode]()

    def run_auto(self):
        self._history['task'] = {'data': self.task, 'id': 0}
        task = self.task
        is_end = False
        attempt = 0
        while not is_end:
            attempt += 1
            response = self.control_step(task)['result']
            task = response
            func_call = self.control_step.history[-1].get('function_call')
            if func_call:
                func_call = ast.literal_eval(func_call)
                self._history[func_call.get('name', 'unknown')] = {'data': response, 'id': attempt}
                if func_call.get('name') == 'end_step':
                    is_end = True
            else:
                self._history['control'] = {'data': response, 'id': attempt}
            if attempt == self.max_attempts:
                is_end = True
        if self._history.get("end_step"):
            self.result = self._history["end_step"]['data']
        else:
            self.result = "Max attempts achieved."
        return self


    def run_static(self):
        self._history['task'] = {'data': self.task, 'id': 0}
        for i, flow_step in enumerate(self.flow):
            step_inputs = dict()
            for arg_name, arg_value in flow_step["inputs"].items():
                step_inputs[arg_name] = self._history[arg_value[1:]]['data'] if "$" in arg_value else arg_value
            step = self.scenario.steps[flow_step["name"]]
            response = step(**step_inputs)
            self._history[flow_step["name"]] = {'data': response, 'id': i+1}
        self.result = self._history["end_step"]['data']
        return self
